/*
 * Copyright (c) 2005 Silicon Graphics, Inc.
 * All rights reserved.
 *
 *   Michael A. Raymond <mraymond@sgi.com>
 */

/*
 *  This file is part of the intr_rate application.
 *
 *  intr_rate is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  intr_rate is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with intr_rate; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <stdlib.h>
#include <uli.h>
#include <sys/mman.h>
#include <sys/types.h>

#define SAMPLES 10000

u_int64_t times[SAMPLES];
int samples;
int max_samples = SAMPLES;
int uli_data = 0; /* Specialize this later */
void * uli_intr;

/*
 * This is the ULI interrupt handler.  It runs every time our interrupt
 * of interest fires.  It records when the interrupt occured in a global
 * buffer.  After N runs it wakes up the main program.
 */
static void
uli_handler(void * arg)
{
	/* Init run, just quit */
	if (arg == NULL) {
		return;
	}

	/*
	 * Already taken N samples, nothing more to do.  The main application
	 * will be killing us off soon.
	 */
	if (samples >= max_samples) {
		return;
	}

	/* Get the current time */
	asm volatile("mov %0=ar.itc" : "=r" (times[samples]));

	/* If we've taken N samples, wake the application back up */
	samples++;
	if (samples == max_samples) {
		ULI_wakeup(uli_intr, 0);
	}
}

/*
 * Do simple analysis on the interarrival times of the interrupts
 */
static void
analyze_intrs(void)
{
	u_int64_t max_time, min_time;
	double sum;
	int i;

	max_time = min_time = sum = times[1] - times[0];

	/* Loop through the samples gathering the results */
	for (i = 2; i < max_samples; i++) {
		u_int64_t diff = times[i] - times[i - 1];

		sum += diff;
		if (diff < min_time) {
			min_time = diff;
		}
		if (diff > max_time) {
			max_time = diff;
		}
	}

	/* Print the results */
	printf("All time results are in units of CPU cycle counter ticks\n");
	printf("Min time: %ld\n", min_time);
	printf("Max time: %ld\n", max_time);
	printf("Avg time: %lf\n", sum / (max_samples - 1));

	return;
}

/*
 * Print a help message
 */
static void
print_help(char * name)
{
	fprintf(stderr, "Execute as %s I C S\n", name);
	fprintf(stderr, "\tI: IRQ to use\n");
	fprintf(stderr, "\tC: CPU to use\n");
	fprintf(stderr, "\tS: # of samples to take\n");
}

int
main(int argc, char * argv[])
{
	int irq, cpu;

	if (argc < 4) {
		print_help(argv[0]);
		return 1;
	}

	/* Get the input arguments */
	irq = atoi(argv[1]);
	cpu = atoi(argv[2]);
	max_samples = atoi(argv[3]);
	if (max_samples < 1 || max_samples > SAMPLES) {
		fprintf(stderr, "Must take between 0 and %d samples\n",
			 SAMPLES);
		return 1;
	}

	/* Lock memory */
	if (mlockall(MCL_CURRENT | MCL_FUTURE) < 0) {
		perror("mlockall");
		return 1;
	}

	/* Reference our interrupt handler functions ahead of time */
	uli_handler(NULL);
	ULI_wakeup(NULL, -1);

	/* Register the ULI handler */
	if ((uli_intr = ULI_register_irq(cpu, uli_handler, &uli_data, 1,
					 NULL, 0, irq)) == NULL) {
		perror("ULI_register_irq");
		return 1;
	}

	/* Wait for the handler to take N samples */
	if (ULI_sleep(uli_intr, 0) < 0) {
		perror("ULI_sleep");
		return 1;
	}

	/* We can turn off the handler now */
	if (ULI_destroy(uli_intr) < 0) {
		perror("ULI_destroy");
		return 1;
	}
	
	/* Analyze the results */
	analyze_intrs();

	return 0;
}
